from pathlib import Path
import collections
import logging
import os
import requests
from . import config


# Find the primary extension which we will use
def find_extension(path_to_check):
    EXTENSIONS = ['mp3', 'm4a', 'm4b']

    for EXT in EXTENSIONS:
        if collections.Counter(
            p.suffix for p in Path(path_to_check)
                .resolve().glob(f'*.{EXT}')
                ):
            extension_to_use = EXT
            list_of_files = os.listdir(Path(path_to_check))
            # Case for single file in a folder
            if sum(
                x.endswith(f'.{extension_to_use}')
                for x in list_of_files
            ) == 1:
                for m4b_file in (
                    Path(path_to_check)
                        .glob(f'*.{extension_to_use}')):
                    logging.debug(
                        f"Adjusted input for {path_to_check}"
                        f" to use single m4b file")
                    path_to_use = m4b_file
                num_of_files = 1
            else:
                num_of_files = sum(
                    x.endswith(f'.{extension_to_use}')
                    for x in list_of_files
                    )
                path_to_use = path_to_check
            return path_to_use, extension_to_use, num_of_files


# Validate path, check if it's a directory or a file
def get_directory(input_take):
    # Check if input is a dir
    if Path(input_take).is_dir():
        # Check if input has multiple subdirs
        num_of_subdirs = len(next(os.walk(input_take))[1])
        if num_of_subdirs >= 1:
            logging.info(
                f"Found multiple ({num_of_subdirs}) subdirs, "
                f"using those as input (multi-disc)"
            )
            path_to_use = input_take
            extension_to_use = None
            num_of_files = num_of_subdirs
        else:
            for dirpath, dirnames, files in os.walk(input_take):
                return_find_ext = find_extension(dirpath)
                # Return error if no supported file extensions
                if not return_find_ext:
                    return None
                path_to_use = return_find_ext[0]
                extension_to_use = return_find_ext[1]
                num_of_files = return_find_ext[2]

    # Check if input is a file
    elif Path(input_take).is_file():
        path_to_use = input_take
        extension_to_use_PRE = path_to_use.suffix
        extension_to_use = Path(extension_to_use_PRE).stem.split('.')[1]
        num_of_files = 1

    # Handle not a file or dir
    else:
        logging.error("File not accessible or valid")
        return None

    logging.debug(f"Final input path is: {path_to_use}")
    logging.debug(f"Extension is: {extension_to_use}")
    logging.debug(f"Number of files: {num_of_files}")
    return Path(path_to_use), extension_to_use, num_of_files


# Checks that asin is the expected length, then cheks for http code 200
def validate_asin(asin):
    if len(asin) == 10:
        # Check ASIN http response
        check = requests.get(f"{config.api_url}/{asin}")
        # If either good http response or valid api response
        if check.status_code == 200:
            logging.info(f"Validated ASIN: {asin}")
        else:
            raise ValueError(f"HTTP error {check.status_code}")

        return check.status_code

    raise ValueError("Invalid ASIN length")



from pathlib import Path
import argparse
import logging
import os
import sys
# Local imports
from . import audible_helper, helpers, m4b_helper


def run_all(inputs):
    print('-' * 50)
    print(f"Working on: {inputs}")
    print('-' * 50)
    # Validate path, check if it's a directory or a file
    # This will also run find_extension to determine relevant filetype
    input_data = helpers.get_directory(inputs)

    # Validate ASIN input
    while True:
        try:
            asin = input("Audiobook ASIN: ")
            helpers.validate_asin(asin)
            break
        except Exception as e:
            print(str(e))

    # Create BookData object from asin response
    aud = audible_helper.BookData(asin)
    metadata = aud.fetch_api_data()
    chapters = aud.get_chapters()

    # Process metadata and run components to merge files
    m4b = m4b_helper.M4bMerge(input_data, metadata, chapters)
    m4b.run_merge()


def main():
    parser = argparse.ArgumentParser(
        description='m4bmerge cli'
    )
    parser.add_argument(
        "-i", "--inputs",
        help="Input paths to process",
        nargs='+',
        required='--login' not in sys.argv,
        type=Path
    )
    parser.add_argument(
        "--log_level",
        help="Set logging level"
    )
    parser.add_argument(
        "--login",
        default=False,
        action="store_true",
        help="Log in to Audible"
    )

    args = parser.parse_args()

    # Run login helper
    if args.login:
        aud = audible_helper.AudibleAuth()
        run_auth = aud.handle_auth()
        if not run_auth:
            print("Logged in")
    # Get log level from system or input
    if args.log_level:
        numeric_level = getattr(logging, args.log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % args.log_level)
        logging.basicConfig(level=numeric_level)
    else:
        logging.basicConfig(level=os.environ.get("LOG_LEVEL", "INFO"))
    # Run through inputs
    if args.inputs:
        for inputs in args.inputs:
            if inputs.exists():
                run_all(inputs)
            else:
                logging.error(f"Input \"{inputs}\" does not exist")

        print('-' * 25)
        print(f"Done processing {len(args.inputs)} inputs")
        print('-' * 25)


# Only run call if using CLI directly
if __name__ == "__main__":
    main()



from datetime import datetime
import logging
import math
import os
import requests
import shutil
import subprocess
from pathlib import Path
from pathvalidate import sanitize_filename
from pydub.utils import mediainfo
# Local imports
from . import config, helpers


class M4bMerge:
    def __init__(self, input_data, metadata, chapters=None):
        self.input_path = input_data[0]
        self.input_extension = input_data[1]
        self.num_of_files = input_data[2]
        self.metadata = metadata
        self.chapters = chapters

    def download_cover(self):
        if 'image' in self.metadata:
            # Request to image URL
            cover_request = requests.get(self.metadata['image'])
            # Verify image exists
            if cover_request.status_code == 200:
                # Path to write image to
                self.cover_path = f"{self.input_path}_cover.jpg"
                # Write image
                with open(self.cover_path, 'wb') as f:
                    f.write(cover_request.content)
            else:
                logging.error("Couldn't download Audible cover")
        else:
            logging.warning("No cover image available from Audible")

    def prepare_data(self):
        # Metadata variables
        # Only use subtitle in case of metadata, not file name
        if 'subtitle' in self.metadata:
            base_title = self.metadata['title']
            base_subtitle = self.metadata['subtitle']
            self.title = f"{base_title} - {base_subtitle}"
        else:
            self.title = self.metadata['title']
        # no subtitle for file name
        self.path_title = self.metadata['title']

        # Only use first author/narrator for file names;
        self.path_author = self.metadata['authors'][0]['name']

        # For embedded, use all authors/narrators
        author_name_arr = []
        for authors in self.metadata['authors']:
            author_name_arr.append(authors['name'])
        self.author = ', '.join(author_name_arr)

        narrator_name_arr = []
        for narrators in self.metadata['narrators']:
            narrator_name_arr.append(narrators['name'])
        self.narrator = ', '.join(narrator_name_arr)

        if 'seriesPrimary' in self.metadata:
            self.series = self.metadata['seriesPrimary']['name']
            if 'position' in self.metadata['seriesPrimary']:
                self.series_position = (
                    self.metadata['seriesPrimary']['position']
                )
            else:
                self.series_position = None
        else:
            self.series = None

        self.summary = self.metadata['description']

        # Convert date string into datetime object
        dateObj = datetime.strptime(
            self.metadata['releaseDate'], '%Y-%m-%dT%H:%M:%S.%fZ'
        )
        self.year = dateObj.year

        self.genre = None
        if 'genres' in self.metadata:
            genre_names = []
            for g in self.metadata['genres']:
                genre_names.append(g['name'])
            self.genre = '/'.join(genre_names)

        # Use format type for comment
        if 'formatType' in self.metadata:
            self.comment = self.metadata['formatType'].capitalize()

    def prepare_command_args(self):
        self.book_output = (
            f"{config.output}/{sanitize_filename(self.path_author)}/"
            f"{sanitize_filename(self.path_title)}"
        )
        self.file_title = sanitize_filename(self.title)

        # Download cover image
        self.download_cover()
        ##

        # Make necessary directories
        # Final output folder
        Path(self.book_output).mkdir(
            parents=True,
            exist_ok=True
        )

        # Folder to move original input into
        Path(config.junk_dir).mkdir(
            parents=True,
            exist_ok=True
        )

        # Array for argument use
        # main metadata args
        self.metadata_args = [
            f"--name={self.title}",
            f"--album={self.path_title}",
            f"--artist={self.narrator}",
            f"--albumartist={self.author}",
            f"--year={self.year}",
            f"--description={self.summary}",
        ]

        # Append series to metadata if it exists
        if self.series:
            self.metadata_args.append(f"--series={self.series}")
            if self.series_position:
                self.metadata_args.append(
                    f"--series-part={self.series_position}"
                )

        if self.genre:
            self.metadata_args.append(
                f"--genre={self.genre}"
            )

        if self.comment:
            self.metadata_args.append(
                f"--comment={self.comment}"
            )

        if self.cover_path:
            self.metadata_args.append(
                f"--cover={self.cover_path}"
            )

        # args for merge  process
        self.processing_args = [
            '--force',
            '--no-chapter-reindexing',
            '--no-cleanup',
            '--ignore-source-tags',
            f'--jobs={config.num_cpus}'
        ]

        # Set logging level of m4b-tool depending upon log_level
        level = logging.root.level
        if level == logging.INFO:
            self.processing_args.append('-v')
        elif level == logging.DEBUG:
            self.processing_args.append('-vvv')

    def find_bitrate(self, file_input):
        # Divide bitrate by 1k, round up,
        # and return back to 1k divisible for round number.
        target_bitrate = math.ceil(
            int(mediainfo(file_input)['bit_rate']) / 1000
        ) * 1000
        logging.info(f"Source bitrate: {target_bitrate}")

        return target_bitrate

    def find_samplerate(self, file_input):
        target_samplerate = int(
            mediainfo(
                file_input
            )['sample_rate']
        )
        logging.info(f"Source samplerate: {target_samplerate}")

        return target_samplerate

    def run_merge(self):
        # Prepare metadata commands
        self.prepare_data()

        # Prepare arguments for m4b-tool commands
        self.prepare_command_args()

        # Handle multiple input files in a folder
        if ((self.input_path.is_dir() and self.num_of_files > 1)
                or self.input_extension is None):
            self.merge_multiple_files()
        # Handle single m4b or m4a input file
        elif ((self.input_path.is_file()) and
                (self.input_extension == "m4b" or
                    self.input_extension == "m4a")):
            self.merge_single_aac()
        # Handle single mp3 input file
        elif self.input_path.is_file() and self.input_extension == "mp3":
            self.merge_single_mp3()
        # If none of the above are true, log error
        else:
            logging.error(
                f"Couldn't determine input type/extension for"
                f" {self.file_title}")

    def merge_multiple_files(self):
        logging.info("Processing multiple files in a dir...")

        first_file = self.find_file_for_mediainfo()

        # Mediainfo data
        target_bitrate = self.find_bitrate(first_file)
        target_samplerate = self.find_samplerate(first_file)
        ##

        args = [
            config.m4b_tool_bin,
            'merge',
            f"--output-file={self.book_output}/{self.file_title}.m4b"
        ]

        # Add in main metadata and merge args
        args.extend(self.metadata_args)
        args.extend(self.processing_args)

        if self.input_extension == "m4b" or self.input_extension == "m4a":
            logging.info(
                f"Multiple {self.input_extension} files, not converting"
            )
            args.append("--no-conversion")
        else:
            args.append(f"--audio-bitrate={target_bitrate}")
            args.append(f"--audio-samplerate={target_samplerate}")

        # Append input path
        args.append(self.input_path)

        # m4b command with passed args
        logging.debug(f"M4B command: {args}")
        subprocess.call(args, shell=False)

        # Move obsolete input to processed folder
        self.move_completed_input()
        # Process chapters
        self.fix_chapters()

    def merge_single_aac(self):
        logging.info(f"Processing single {self.input_extension} input...")

        args = [
            config.m4b_tool_bin,
            'meta',
            '--ignore-source-tags',
            (f"{self.input_path.parent}/"
                f"{self.input_path.stem}.new.m4b")
        ]
        # Add in main metadata args
        args.extend(self.metadata_args)

        # make backup file
        shutil.copy(
            self.input_path,
            f"{self.input_path.parent}/{self.input_path.stem}.new.m4b"
            )

        # m4b command with passed args
        logging.debug(f"M4B command: {args}")
        subprocess.call(args, shell=False)

        # Move completed file
        shutil.move(
            f"{self.input_path.parent}/{self.input_path.stem}.new.m4b",
            f"{self.book_output}/{self.file_title}.m4b"
        )

        # Move obsolete input to processed folder
        self.move_completed_input()
        # Process chapters
        self.fix_chapters()

    def merge_single_mp3(self):
        logging.info(f"Processing single {self.input_extension} input...")

        # Mediainfo data
        target_bitrate = self.find_bitrate(self.input_path)
        target_samplerate = self.find_samplerate(self.input_path)
        ##

        args = [
            config.m4b_tool_bin,
            'merge',
            f"--output-file={self.book_output}/{self.file_title}.m4b",
            f"--audio-bitrate={target_bitrate}",
            f"--audio-samplerate={target_samplerate}"
        ]
        # Add in main metadata and merge args
        args.extend(self.metadata_args)
        args.extend(self.processing_args)

        # Append input path
        args.append(self.input_path)

        # m4b command with passed args
        logging.debug(f"M4B command: {args}")
        subprocess.call(args, shell=False)

        # Move obsolete input to processed folder
        self.move_completed_input()
        # Process chapters
        self.fix_chapters()

    def find_file_for_mediainfo(self):
        # If multi-disc, find the extension
        if not self.input_extension:
            selected_input = self.find_multi_disc_extension()
        else:
            selected_input = self.input_path

        selected_input_glob = Path(selected_input).glob('**/*')

        # Find first file with our extension, to check rates against
        for file in sorted(selected_input_glob):
            if file.suffix == f".{self.input_extension}":
                first_file = file
                break

        logging.debug(f"Got file to run mediainfo on: {first_file}")
        return first_file

    def find_multi_disc_extension(self):
        # Get all directories
        input_path_glob = Path(self.input_path).glob('**/*')
        # Find first directory in sort
        i = 0
        sorted_multi = sorted(input_path_glob)
        while not sorted_multi[i].is_dir():
            logging.debug("Looking for first dir in multi-dir...")
            i += 1
        selected_input = sorted_multi[i]
        logging.debug(
            f"Result was #{i+1} for first dir: {selected_input}"
        )
        # Now that first sorted directory was found, find it's primary ext
        self.input_extension = helpers.find_extension(
            selected_input)[1]
        logging.debug(
            (f"Guessed multi-disc extension to be:"
                f" {self.input_extension}")
        )
        return selected_input

    def fix_chapters(self):
        chapter_file = f"{self.book_output}/{self.file_title}.chapters.txt"
        m4b_to_modify = f"{self.book_output}/{self.file_title}.m4b"

        # Use audible chapters if they exist and this isn't an mp3
        if (self.chapters and self.input_extension != "mp3") or (
            self.num_of_files == 1 and self.input_extension == "mp3"
        ):
            logging.info("Using chapter data from Audible")
            new_file_content = ('\n'.join(self.chapters))
        # Else fix formatting of existing chapters
        else:
            logging.info("Using existing chapter data")
            new_file_content = ""
            with open(chapter_file) as f:
                # Store and then skip past total length section
                for line in f:
                    if "# total-length" in line.strip():
                        new_file_content += line.strip() + "\n"
                        break
                # Iterate over rest of the file
                counter = 0
                for line in f:
                    stripped_line = line.strip()
                    counter += 1
                    new_line = (
                        (stripped_line[0:13]) +
                        f'Chapter {"{0:0=2d}".format(counter)}'
                        )
                    new_file_content += new_line + "\n"

        with open(chapter_file, 'w') as f:
            f.write(new_file_content)

        # Apply chapter arguments/command
        args = [
            config.mp4chaps_bin,
            '-z',
            m4b_to_modify,
        ]

        # Check that chapter file is valid
        # Generally only an issue because of:
        # https://github.com/sandreas/m4b-tool/issues/141
        if os.path.getsize(chapter_file) == 0:
            logging.error("Chapter file is empty, attempting to correct")
            args.append('-r')
        else:
            logging.info("Applying chapters to m4b...")
            args.append('-i')

        # Set logging level of m4bchaps depending upon log_level
        if logging.root.level == logging.DEBUG:
            args.append('-v')
        elif logging.root.level >= logging.WARNING:
            args.append('-q')

        # Apply fixed chapters to file
        subprocess.call(args, shell=False)

    def move_completed_input(self):
        # Cleanup cover file
        if self.cover_path:
            os.remove(self.cover_path)
        # Move obsolete input to processed folder
        if Path(self.input_path.parent, 'done') == config.junk_dir:
            logging.debug("Junk dir is direct parent")
            move_dir = Path(self.input_path)
        elif Path(self.input_path.parents[1], 'done') == config.junk_dir:
            logging.debug("Junk dir is double parent")
            move_dir = Path(self.input_path.parent)
        else:
            return logging.warning("Couldn't find junk dir relative to input")

        dest = Path(config.junk_dir, move_dir.name)
        try:
            move_dir.replace(dest)
        except OSError:
            logging.warning("Couldn't move input to complete dir")



import logging
import requests
from datetime import timedelta
from . import config


# Checks validity of asin, then gathers json response into a return object
class BookData:
    def __init__(self, asin):
        self.asin = asin

    # Convert MS to timestamp format hh:mm:ss.ms
    def ms_to_timestamp(self, input_duration):
        conversion = timedelta(milliseconds=input_duration)
        # Hacky fix for timedelta showing days past 24hr
        # Test if we've gone into days
        if "day" in str(conversion):
            # Find hour after days string
            hour = int(str(conversion).split(", ")[1].split(":")[0])
            # Find int number of days
            day = int(str(conversion).split(" ")[0])
            # Add 24 hours for each day
            real_hours = (hour+(day*24))
            # Get the rest of the time string based on colon split
            remainder = str(conversion).split(", ")[1].split(":", 1)[1]
            timestamp = f"{real_hours}:{remainder}"
        else:
            timestamp = str(conversion)

        # Remove trailing 000 if it makes ms 6 places long
        if "." in timestamp:
            split_timestamp = timestamp.split(".")
            prefix = split_timestamp[0]
            suffix = split_timestamp[1]
            if len(suffix) > 3:
                suffix = suffix.rstrip("000")
            return prefix + '.' + suffix

        return timestamp + '.' + '000'

    def get_chapters(self):
        # Select chapter data from json response
        chapter_info = self.metadata_dict['chapterInfo']

        # Only use Audible chapters if tagged as accurate
        if chapter_info['isAccurate'] is True:
            chapter_output = []
            # Append total runtime to the top of file
            total_len = self.ms_to_timestamp(chapter_info['runtimeLengthMs'])
            chapter_output.append(
                (
                    "# total-length"
                    f" {total_len}"
                )
            )

            # Append each chapter to array
            for chapter in chapter_info['chapters']:
                chap_start = self.ms_to_timestamp(chapter['startOffsetMs'])
                # Starting chapter title data
                chapter_title = chapter['title']
                chapter_output.append(
                    (
                        f"{chap_start}"
                        f" {chapter_title}"
                    )
                )
        else:
            logging.warning(
                "Not using Audible chapters as they aren't tagged as accurate"
            )
            chapter_output = None

        return chapter_output

    def fetch_api_data(self):
        # metadata dictionary
        api_call = requests.get(f"{config.api_url}/{self.asin}")
        self.metadata_dict = api_call.json()
        return self.metadata_dict

    def check_asin_sku(self):
        # Login or register as needed
        self.auth.handle_auth()
        aud_json = self.auth.client.get(
            f"catalog/products/{self.asin}",
            params={
                "response_groups": "sku",
                "asins": self.asin
            }
        )
        if 'sku' in aud_json['product']:
            return True
        return None



from pathlib import Path
import os
import shutil
from appdirs import user_config_dir

# User editable variables
# API url
api_url = "https://api.audnex.us/books"

# output directory for cleaned metadata/folder structure
# leaving blank uses /output for docker or $USER/output for anything else
output = ""

# Number of cpus to use for jobs
cpus_to_use = ""
#

# Set defaults if no user changes
# Setup output folder defaults
if not output:
    # If using docker, default to /output folder, else $USER/output
    if Path('/output').is_dir():
        output = Path('/output')
    else:
        default_output = Path.home()
        output = Path(f"{default_output}/output")

# If using docker, default to /input/done folder, else $USER/input/done
if Path('/input').is_dir():
    junk_dir = Path('/input/done')
else:
    default_input = Path.home()
    junk_dir = Path(f"{default_input}/input/done")

# Available CPU cores to use
if not cpus_to_use:
    num_cpus = os.cpu_count()
else:
    num_cpus = cpus_to_use

# config section for docker
if Path('/config').is_dir():
    config_path = Path('/config')
else:
    appname = "m4b-merge"
    config_path = Path(user_config_dir(appname))
    Path(config_path).mkdir(
        parents=True,
        exist_ok=True
    )

# Find path to m4b-tool binary
# Check that binary actually exists
m4b_tool_bin = shutil.which('m4b-tool')
if not m4b_tool_bin:
    raise SystemExit(
        'Error: Cannot find m4b-tool binary.'
        )

# Test existence of mp4chaps
mp4chaps_bin = shutil.which('mp4chaps')
if not mp4chaps_bin:
    raise SystemExit(
        'Error: Cannot find mp4chaps binary.'
        )
###



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='m4b-merge',
    version='0.4',
    url='https://github.com/djdembeck/m4b-merge',
    description=(
                "A tool to standardize audiobook files"
                " into single m4b files with clean metadata"
                " based on Audible and chapters."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='David Dembeck',
    author_email='dembeck.david@protonmail.com',
    project_urls={
        "Bug Tracker": "https://github.com/djdembeck/m4b-merge/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'appdirs>=1.4.4',
        'pathvalidate>=2.4.1',
        'pydub>=0.25.1',
        'requests>=2.24.0',
    ],
    package_dir={"": "src"},
    packages=["m4b_merge"],
    python_requires=">=3.6",
    entry_points={'console_scripts': ['m4b-merge = m4b_merge.__main__:main']}
)
